import stat
from typing import List

from pyscaffold.actions import Action, ActionParams, ScaffoldOpts, Structure
from pyscaffold.extensions import Extension, include
from pyscaffold.extensions.no_skeleton import NoSkeleton
from pyscaffold.extensions.pre_commit import PreCommit
from pyscaffold.identification import get_id
from pyscaffold.operations import add_permissions, no_overwrite
from pyscaffold.structure import ensure, merge

from pyscaffoldext.markdown.extension import Markdown, convert_files

from .templates import readme_md, template

NO_OVERWRITE = no_overwrite()


class DSProject(Extension):
    """Template for data-science projects"""

    name = "dsproject"

    def augment_cli(self, parser):
        """Augments the command-line interface parser

        A command line argument ``--FLAG`` where FLAG=``self.name`` is added
        which appends ``self.activate`` to the list of extensions. As help
        text the docstring of the extension class is used.
        In most cases this method does not need to be overwritten.

        Args:
            parser: current parser object
        """

        parser.add_argument(
            self.flag,
            help=self.help_text,
            nargs=0,
            dest="extensions",
            action=include(NoSkeleton(), PreCommit(), self),
        )
        return self

    def activate(self, actions: List[Action]) -> List[Action]:
        actions = Markdown().activate(actions)
        # ^  Wrapping the Markdown extension is more reliable then including it via CLI.
        #    This way we can trust the activation order for registering actions,
        #    and the Python API is guaranteed to work, even if the user does not include
        #    Markdown in the list of extensions.
        actions = self.register(actions, add_dsproject, after="define_structure")
        actions = self.register(actions, replace_readme, after=get_id(convert_files))
        return actions


def add_dsproject(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Adds basic module for custom extension
    See :obj:`pyscaffold.actions.Action`
    """
    gitignore_all = (template("gitignore_all"), NO_OVERWRITE)

    files: Structure = {
        "configs": {".gitignore": ("", NO_OVERWRITE)},
        "data": {
            ".gitignore": (template("gitignore_data"), NO_OVERWRITE),
            **{
                folder: {".gitignore": gitignore_all}
                for folder in ("external", "interim", "preprocessed", "raw")
            },
        },
        "environment.yml": (template("environment_yml"), NO_OVERWRITE),
        "models": {".gitignore": gitignore_all},
        "notebooks": {"template.ipynb": (template("template_ipynb"), NO_OVERWRITE)},
        "references": {".gitignore": ("", NO_OVERWRITE)},
        "reports": {"figures": {".gitignore": ("", NO_OVERWRITE)}},
        "scripts": {
            "train_model.py": (
                template("train_model_py"),
                add_permissions(stat.S_IXUSR, NO_OVERWRITE),
            )
        },
    }

    return merge(struct, files), opts


def replace_readme(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Replace the readme.md of the markdown extension by our own
    See :obj:`pyscaffold.actions.Action`
    """
    return ensure(struct, "README.md", readme_md, NO_OVERWRITE), opts



"""Place for fixtures and configuration that will be used in most of the tests.
A nice option is to put your ``autouse`` fixtures here.
Functions that can be imported and re-used are more suitable for the ``helpers`` file.
"""
import os

import pytest

from .helpers import rmpath, uniqstr


@pytest.fixture
def tmpfolder(tmp_path):
    old_path = os.getcwd()
    new_path = tmp_path / uniqstr()
    new_path.mkdir(parents=True, exist_ok=True)
    os.chdir(str(new_path))
    try:
        yield new_path
    finally:
        os.chdir(old_path)
        rmpath(new_path)



from functools import partial

from pyscaffold.templates import ScaffoldOpts, get_template

from .. import __version__ as dsproject_version

template = partial(get_template, relative_to=__name__)


def readme_md(opts: ScaffoldOpts) -> str:
    """Adds a basic README.md

    Args:
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        file content as string
    """
    opts["pkg"] = opts["package"].ljust(19)
    opts["dsproject_version"] = dsproject_version
    return template("readme_md").safe_substitute(opts)



from pathlib import Path

import pytest
from pyscaffold import cli
from pyscaffold.file_system import chdir

from pyscaffoldext.dsproject.extension import DSProject

from .helpers import run_common_tasks

EXT_FLAG = DSProject().flag

# If you need to check logs with caplog, have a look on
# pyscaffoldext-custom-extension's tests/conftest.py file and the
# `isolated_logger` fixture.

EXPECTED_DIRS = [
    "configs",
    "data/external",
    "data/interim",
    "data/preprocessed",
    "data/raw",
    "models",
    "reports/figures",
]

EXPECTED_FILES = [
    "README.md",
    "environment.yml",
    "notebooks/template.ipynb",
    "scripts/train_model.py",
    "src/my_package/__init__.py",
]


def test_add_custom_extension(tmpfolder):
    args = ["--no-config", "-vv", EXT_FLAG, "my_project", "-p", "my_package"]
    # --no-config: avoid extra config from dev's machine interference
    cli.main(args)

    with chdir("my_project"):
        assert not Path("README.rst").exists()
        for path in EXPECTED_DIRS:
            assert Path(path).is_dir()
        for path in EXPECTED_FILES:
            assert Path(path).is_file()


def test_add_custom_extension_and_pretend(tmpfolder):
    args = [
        "--no-config",
        "--pretend",
        "-vv",
        EXT_FLAG,
        "my_project",
        "-p",
        "my_package",
    ]
    # --no-config: avoid extra config from dev's machine interference
    cli.main(args)

    assert not Path("my_project").exists()


@pytest.mark.slow
@pytest.mark.system
def test_generated_extension(tmpfolder):
    args = ["--no-config", "-vv", "--venv", "--pre-commit", EXT_FLAG, "my_project"]
    # --no-config: avoid extra config from dev's machine interference
    # --venv: generate a venv so we can install the resulting project
    # --pre-commit: ensure generated files respect repository conventions
    cli.main(args)

    with chdir("my_project"):
        # Testing a project generated by the custom extension
        run_common_tasks(tests=False)
